/**
 * Copyright 2008 Bluestem Software LLC.  All Rights Reserved.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

package org.bluestemsoftware.open.eoa.ext.binding.http.dfault.mgmt;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.management.JMException;
import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.bluestemsoftware.open.eoa.commons.util.DOMUtils;
import org.bluestemsoftware.open.eoa.ext.binding.http.dfault.HTTPBindingImpl;
import org.bluestemsoftware.open.eoa.ext.binding.http.dfault.mgmt.jmx.BindingMXBean;
import org.bluestemsoftware.open.eoa.ext.binding.http.dfault.util.Constants;
import org.bluestemsoftware.open.eoa.ext.binding.http.dfault.util.DOMSerializer;
import org.bluestemsoftware.open.eoa.ext.binding.http.dfault.util.HTTPBindingProviderConfiguration;
import org.bluestemsoftware.specification.eoa.Resource;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentName;
import org.bluestemsoftware.specification.eoa.ext.ExecutableException;
import org.bluestemsoftware.specification.eoa.ext.ExtensionFactoryContext;
import org.bluestemsoftware.specification.eoa.ext.binding.http.HTTPBinding;
import org.bluestemsoftware.specification.eoa.ext.binding.http.HTTPBindingException;
import org.bluestemsoftware.specification.eoa.ext.binding.http.rt.HTTPBindingRT;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.InputSource;

/**
 * Encapsulates feature management which includes implementation of administrative operations
 * which MAY be invoked at runtime.
 */

public class Binding implements BindingMXBean {

    private HTTPBindingImpl up;
    private HTTPBindingProviderConfiguration hbc;

    public Binding(ExtensionFactoryContext factoryContext, HTTPBindingRT consumer, HTTPBindingImpl up) throws HTTPBindingException {
        
        this.up = up;
        
        org.bluestemsoftware.specification.eoa.component.binding.Binding binding = null;
        binding = consumer.getRuntimeProvidable();
        
        Element configuration = null;
        File configurationFile = new File(consumer.getExtensionEtcDir(), "config.xml");
        if (configurationFile.exists()) {
            try {
                DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
                factory.setNamespaceAware(true);
                DocumentBuilder documentBuilder = factory.newDocumentBuilder();
                Document configurationDoc = documentBuilder.parse(configurationFile);
                configuration = configurationDoc.getDocumentElement();
            } catch (Exception ex) {
                throw new HTTPBindingException("Error loading configuration. " + ex);
            }
        }

        // if configuration is null, retrieve default configuration contained within
        // deployment and persist it

        if (configuration == null) {
            Writer writer = null;
            try {
                Resource resource = factoryContext.getResource(Constants.DEFAULT_CONFIGURATION_LOC);
                InputSource inputSource = new InputSource(resource.getInputStream());
                inputSource.setSystemId(Constants.DEFAULT_CONFIGURATION_LOC);
                DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
                factory.setNamespaceAware(true);
                DocumentBuilder documentBuilder = factory.newDocumentBuilder();
                documentBuilder.setEntityResolver(factoryContext);
                configuration = documentBuilder.parse(inputSource).getDocumentElement();
                writer = new FileWriter(configurationFile);
                DOMSerializer.serializeNode(configuration, writer, "UTF-8", true);
            } catch (Exception ex) {
                throw new HTTPBindingException("Error retrieving default configuration. " + ex);
            } finally {
                if (writer != null) {
                    try {
                        writer.close();
                    } catch (IOException ignore) {
                    }
                }
            }
        }
        
        hbc = new HTTPBindingProviderConfiguration();
        
        QName childName = new QName(HTTPBinding.TYPE, "operations");
        Element operationsElement = DOMUtils.getChildElement(configuration, childName);
        if (operationsElement != null) {
            childName = new QName(HTTPBinding.TYPE, "operation");
            Map<QName, Integer> requestorTimeout = new HashMap<QName, Integer>();
            List<Element> list = DOMUtils.getChildElements(operationsElement, childName);
            for (Element operationElement : list) {

                String prefixedName = operationElement.getAttribute("name");
                QName operationName = ComponentName.valueOf(operationElement, prefixedName);
                if (binding.getOperation(operationName) == null) {
                    throw new HTTPBindingException("Error configuring runtime provider "
                            + binding.getName()
                            + ". Operation element defined within configuration for operation "
                            + operationName
                            + ", but corresponding binding defines no such operation.");
                }

                childName = new QName(HTTPBinding.TYPE, "requestorTimeout");
                Element timeoutElement = DOMUtils.getChildElement(operationElement, childName);
                if (timeoutElement != null) {
                    Integer timeout = Integer.valueOf(DOMUtils.getText(timeoutElement));
                    requestorTimeout.put(operationName, timeout);
                }

            }
            hbc.setRequestorTimeout(requestorTimeout);
        }

        childName = new QName(HTTPBinding.TYPE, "dataSourceConnector");
        Element dsElement = DOMUtils.getChildElement(configuration, childName);
        if (dsElement != null) {
            hbc.setDataSourceConnectorName(DOMUtils.getText(dsElement));
        }

        childName = new QName(HTTPBinding.TYPE, "embedStackTraceInFaults");
        Element embedElement = DOMUtils.getChildElement(configuration, childName);
        if (embedElement != null) {
            hbc.setEmbeddStackTraceInFaults(Boolean.parseBoolean(DOMUtils.getText(embedElement)));
        }

        childName = new QName(HTTPBinding.TYPE, "maxPayloadProcessors");
        Element payloadElement = DOMUtils.getChildElement(configuration, childName);
        if (payloadElement != null) {
            int value = Integer.parseInt(DOMUtils.getText(payloadElement));
            hbc.setMaxPayloadProcessors(value);
        }
        
    }
    
    /*
     * TODO: when config is changed via mgmnt ui, we need to modify provider at
     * runtime somehow. currently config is retrieved only at start-up
     */
    public HTTPBindingProviderConfiguration getHTTPBindingProviderConfig() {
        return hbc;
    }

    public void resume() throws JMException {
        try {
            up.spi_resume();
        } catch (ExecutableException se) {
            throw new JMException(se.getMessage());
        }
    }

    public void suspend() throws JMException {
        try {
            up.spi_suspend();
        } catch (ExecutableException se) {
            throw new JMException(se.getMessage());
        }
    }

    public Boolean isSuspended() {
        return up.spi_isSuspended();
    }

}
